#include "sys/UserMemory.h"
#include "memory/Memory.h"
#include "memory/MemoryManager.h"
#include "std/stdlib.h"
#include "std/string.h"
#include "utils/Addresses.h"

struct dynamic_string
{
    char* buf;
    long capacity;
    long size;
};

bool dynamic_expand(dynamic_string* str, long new_capacity)
{
    char* buffer = (char*)krealloc(str->buf, new_capacity);
    if (!buffer) { return false; }
    str->buf = buffer;
    str->capacity = new_capacity;
    return true;
}

bool dynamic_push(dynamic_string* str, char c)
{
    if (str->size == str->capacity)
    {
        if (!dynamic_expand(str, str->capacity + 16)) return false;
    }
    str->buf[str->size] = c;
    str->size++;
    return true;
}

bool dynamic_init(dynamic_string* str)
{
    str->buf = (char*)kmalloc(10);
    if (!str->buf) return false;
    str->capacity = 10;
    str->size = 0;
    return true;
}

bool validate_user_readable_page(uintptr_t address)
{
    if (Memory::is_kernel_address(address)) return false;
    auto rc = VMM::get_flags(address);
    if (rc == (uint64_t)-1) return false;
    if (rc & MAP_USER) return true;
    return false;
}

bool validate_user_writable_page(uintptr_t address)
{
    if (Memory::is_kernel_address(address)) return false;
    auto rc = VMM::get_flags(address);
    if (rc == (uint64_t)-1) return false;
    if (rc & (MAP_USER | MAP_READ_WRITE)) return true;
    return false;
}

Result<char*> strdup_from_user(const char* user_string)
{
    uintptr_t user_ptr = (uintptr_t)user_string;
    auto aligned = round_down_to_nearest_page(user_ptr);
    char* ptr = nullptr;
    uintptr_t index = 0;
    if (aligned != user_ptr) // Otherwise, we already do this check below.
    {
        if (!validate_user_readable_page(aligned)) return nullptr;
        auto result = MemoryManager::get_mapping((void*)VMM::get_physical(aligned), 0);
        if (result.has_error()) return result.release_error();
        ptr = (char*)result.release_value();
        index = user_ptr - aligned;
    }
    dynamic_string str;
    if (!dynamic_init(&str))
    {
        if (ptr) MemoryManager::release_mapping(ptr);
        return {ENOMEM};
    }
    while (true) // FIXME: set a limit for this and fail with ENAMETOOLONG otherwise.
    {
        if (user_ptr % PAGE_SIZE == 0)
        {
            index = 0;
            if (ptr) MemoryManager::release_mapping(ptr);
            if (!validate_user_readable_page(user_ptr))
            {
                kfree(str.buf);
                return {EFAULT};
            }
            auto result = MemoryManager::get_mapping((void*)VMM::get_physical(user_ptr), 0);
            if (result.has_error()) return result.release_error();
            ptr = (char*)result.release_value();
        }
        char c = ptr[index];
        if (!dynamic_push(&str, c))
        {
            MemoryManager::release_mapping(ptr);
            kfree(str.buf);
            return {ENOMEM};
        }
        if (!c) // We reached the null terminator!!
        {
            MemoryManager::release_mapping(ptr);
            return str.buf;
        }
        user_ptr++;
        index++;
    }
}

bool validate_user_read(uintptr_t address, size_t size)
{
    auto aligned = round_down_to_nearest_page(address);
    if (aligned != address) // Otherwise, we already do this check below.
    {
        if (!validate_user_readable_page(aligned)) return false;
    }
    while (size--)
    {
        if (address % PAGE_SIZE == 0)
        {
            if (!validate_user_readable_page(address)) return false;
        }
        address++;
    }
    return true;
}

bool validate_user_write(uintptr_t address, size_t size)
{
    auto aligned = round_down_to_nearest_page(address);
    if (aligned != address) // Otherwise, we already do this check below.
    {
        if (!validate_user_writable_page(aligned)) return false;
    }
    while (size--)
    {
        if (address % PAGE_SIZE == 0)
        {
            if (!validate_user_writable_page(address)) return false;
        }
        address++;
    }
    return true;
}

bool do_copy_from_user(const char* uptr, char* ptr, size_t size)
{
    uintptr_t user_ptr = (uintptr_t)uptr;
    auto aligned = round_down_to_nearest_page(user_ptr);
    char* mapping = nullptr;
    uintptr_t index = 0;
    if (aligned != user_ptr) // Otherwise, we already do this check below.
    {
        if (!validate_user_readable_page(aligned)) return false;
        auto result = MemoryManager::get_mapping((void*)VMM::get_physical(aligned), 0);
        if (result.has_error()) return false; // FIXME: Propagate errors.
        mapping = (char*)result.release_value();
        index = user_ptr - aligned;
    }
    while (size--)
    {
        if (user_ptr % PAGE_SIZE == 0)
        {
            if (mapping) MemoryManager::release_mapping(mapping);
            index = 0;
            if (!validate_user_readable_page(user_ptr)) return false;
            auto result = MemoryManager::get_mapping((void*)VMM::get_physical(user_ptr), 0);
            if (result.has_error()) return false; // FIXME: Propagate errors.
            mapping = (char*)result.release_value();
        }
        *ptr = mapping[index];
        user_ptr++;
        ptr++;
        index++;
    }
    return true;
}

bool do_copy_to_user(char* uptr, const char* ptr, size_t size)
{
    uintptr_t user_ptr = (uintptr_t)uptr;
    auto aligned = round_down_to_nearest_page(user_ptr);
    char* mapping = nullptr;
    uintptr_t index = 0;
    if (aligned != user_ptr) // Otherwise, we already do this check below.
    {
        if (!validate_user_writable_page(aligned)) return false;
        auto result = MemoryManager::get_mapping((void*)VMM::get_physical(aligned));
        if (result.has_error()) return false;
        mapping = (char*)result.release_value();
        index = user_ptr - aligned;
    }
    while (size--)
    {
        if (user_ptr % PAGE_SIZE == 0)
        {
            if (mapping) MemoryManager::release_mapping(mapping);
            index = 0;
            if (!validate_user_writable_page(user_ptr)) return false;
            auto result = MemoryManager::get_mapping((void*)VMM::get_physical(user_ptr), 0);
            if (result.has_error()) return false;
            mapping = (char*)result.release_value();
        }
        mapping[index] = *ptr;
        user_ptr++;
        ptr++;
        index++;
    }
    return true;
}

bool copy_from_user(const void* user_ptr, void* ptr, size_t size)
{
    return do_copy_from_user((const char*)user_ptr, (char*)ptr, size);
}

bool copy_to_user(void* user_ptr, const void* ptr, size_t size)
{
    return do_copy_to_user((char*)user_ptr, (const char*)ptr, size);
}